#!perl
#
# Export Ogre3d Mesh XML file
# author: Timothy A. Graupmann
#
# This script will export an OGRE MESH XML file
#
# Good resources:
#
# C:\Program Files\Luxology\modo 201\Documentation\help\Scripting and Commands.pdf
#
#
#

#
# AUTOFLUSH
#
$| = 1;

#
# DEFAULTS
#
my $g_version = "1.0.0.0";
my $g_filter = "*.xml;*.xml";
my $g_ogreTools = "C:\\OGRETOOLS";

#
# VERSION OUTPUT
#
LuxPrint("Export Ogre3d XML Version: $g_version");

#
# EXAMINE ENVIRONMENT VARIABLES
#
#LuxPrint("Available environment variables:");
#LuxPrint(join ("\n", %ENV));

#
# OVERRIDE FILTER DEFAULT WITH ENVIRONMENT VARIABLE
#
if (defined($ENV{"USERPROFILE"}))
{
  $g_version = $ENV{"USERPROFILE"} . "\\Application Data\\Luxology\\Scripts\\*.mesh;*.xml";
}

#
# CREATE OPTION DIALOG
#
#LuxCommand("layout.create width:[600] height:[400]");
#LuxCommand("attr.formCreate \"Ogre Mesh Import Form\"");
#LuxCommand("attr.addSubform \"Level Of Detail\"");
#return;

#
# EXTRACT THE FILES FROM THE DIALOG
#
my $g_filename = SaveDialog($g_version);

#
# ABORT IF NO FILES WERE SELECTED
#
if (!defined($g_filename) || ($g_filename eq ""))
{
  LuxPrint("User aborted.");
  return;
}

#before starting switch to 3 poly mode
LuxCommand("poly.triple");

# Get the layer info
@FGlayers = lxq("query layerservice layers ? fg");

# Get the verts for each layer
&GetVertsInfo;

# Get Polygon Info
&GetPolyInfo;

# Create mesh xml file
WriteFile($g_filename);

#
# THE END
#
return;




######
###### HELPERS
######

# lxout can only print so much per line
sub LuxPrint
{
  my $luxPrint = shift @_;
  my @luxPrint = split(/\n/, $luxPrint);
  foreach my $line (@luxPrint)
  {
    lxout $line;
    #print "$line\n";
  }
}

sub LuxCommand
{
  my $luxCommand = shift @_;
  lx ($luxCommand);
  LuxPrint ($luxCommand);
}

sub LuxQuery
{
  my $luxQuery = shift @_;
  #LuxPrint ("LuxQuery: [$luxQuery]");
  return lxq ($luxQuery);
}

sub PrintHash
{
  my $hash_ref = shift @_;
  foreach my $key (sort %$hash_ref)
  {
    if ($key ne "")
    {
      LuxPrint ($key);
    }
  }
}

sub SaveDialog
{
  my $filter = shift @_;

  lx("dialog.setup fileSave");
  lx("dialog.title {Export binary Ogre3d mesh file}");
  lx("dialog.filetype text");
  lx("dialog.result {$filter}");
  lx("dialog.open");

  my $filename = LuxQuery("dialog.result ?");

  if(!defined $filename)
  {
    return "";
  }
  
  if (rindex($file, '.mesh.xml') < 0)
  {
    $filename .= '.mesh.xml';
  }
  
  $filename =~ s/\.TXT\.mesh\.xml/\.mesh\.xml/;
  
  return $filename;
}

sub GetVertsInfo
  {
    foreach $layer(@FGlayers)
      {
        lx("select.layer $layer");
        lxq("query layerservice layer.index ? selected");
        @verts = lxq("query layerservice verts ? all");
        
        foreach $vert(@verts)
          {
            $layerhash = "Layer_$layer";
            @vert_position = lxq("query layerservice vert.pos ? $vert");
            $vert_position[0] = sprintf("%.6f", $vert_position[0]);
            $vert_position[0] = "x=\"$vert_position[0]\"";
            $vert_position[1] = sprintf("%.6f", $vert_position[1]);
            $vert_position[1] = "y=\"$vert_position[1]\"";
            $vert_position[2] = sprintf("%.6f", $vert_position[2]);
            $vert_position[2] = "z=\"$vert_position[2]\"";
            $$layerhash{"$vert"} = "@vert_position";
            #lxout("@vert_position");
            push(@$layerhash,"$vert");
          }
            
        foreach $vert(@verts)
          {
            
            $layerhash = "Layer_VertNormals_$layer";
            @vert = lxq("query layerservice vert.normal ? $vert");
            $vert[0] = sprintf("%.6f", $vert[0]);
            $vert[0] = "x=\"$vert[0]\"";
            $vert[1] = sprintf("%.6f", $vert[1]);
            $vert[1] = "y=\"$vert[1]\"";
            $vert[2] = sprintf("%.6f", $vert[2]);
            $vert[2] = "z=\"$vert[2]\"";
            $$layerhash{"$vert"} = "@vert";
            lxout("@vert");
            push(@$layerhash,"$vert");
          }
      }
  }
  
sub GetPolyInfo
  {
    foreach $layer(@FGlayers)
      {
        lx("select.layer $layer");
        lxq("query layerservice layer.index ? selected");
        @polys = lxq("query layerservice polys ? all");
        
        foreach $poly(@polys)
          {
            $layerhash = "LayerPoly_$layer";
            @poly_verts = lxq("query layerservice poly.vertList ? $poly");
            for (my $index = 0; $index < scalar @poly_verts; ++$index)
            {
              $poly_verts[$index] = "v" . ($index+1) . "=\"" . $poly_verts[$index] . "\"";
            }
            $$layerhash{"$poly"} = "@poly_verts";
            #lxout("@poly_verts");
            push(@$layerhash,"$poly");
          }
      }
  }

sub WriteFile
{
  my $filename = shift @_;

  #
  # CREATE THE SAVE FILE
  #
  LuxPrint("Create filename: $filename");

  #
  # CHECK IF SAVE FILE CAN BE OPENED
  #
  my $g_openOkay = "true";
  open (FILEHANDLE, ">$filename") or $g_openOkay = "false";
  if ($g_openOkay eq "false")
  {
    LuxPrint("Unable to open save file. [$filename]");
    return;
  }
  
  print FILEHANDLE "<mesh>\n";
  print FILEHANDLE "\t<submeshes>\n";
  foreach $layer(@FGlayers)
  {
    $layerverts = "Layer_$layer";
    $layervertnormals = "Layer_VertNormals_$layer";
    $layerpolys = "LayerPoly_$layer";
    @layervertkeys = keys %$layerverts;
    @layerpolykeys = keys %$layerpolys;
    
    @layervertkeys = sort(@layervertkeys);
    
    $vertcount = (($#layervertkeys + 1));
    $vertmult = $vertcount*3;
    $polycount = ($#layerpolykeys + 1);
  
    print FILEHANDLE "\t\t<submesh material=\"default\" usesharedvertices=\"false\" use32bitindexes=\"false\" operationtype=\"triangle_list\">\n";
    
    ###
    ### FACES
    ###
    print FILEHANDLE "\t\t\t<faces count=\"$polycount\">\n";
    foreach $poly(@$layerpolys)
    {
      $work = $$layerpolys{"$poly"};
      print FILEHANDLE "\t\t\t\t<face $work/>\n";
    }
    print FILEHANDLE "\t\t\t</faces>\n";
    
    ###
    ### GEOMETRY
    ###
    print FILEHANDLE "\t\t\t<geometry vertexcount=\"$vertcount\">\n";
    
    ###
    ### VERTEX BUFFER POSITIONS
    ###
    print FILEHANDLE "\t\t\t\t<vertexbuffer positions=\"true\">\n";
    foreach $vert(@$layerverts)
    {
      $work = $$layerverts{"$vert"};
      print FILEHANDLE "\t\t\t\t\t<vertex>\n";
      print FILEHANDLE "\t\t\t\t\t\t<position $work />\n";
      print FILEHANDLE "\t\t\t\t\t</vertex>\n";
    }
    print FILEHANDLE "\t\t\t\t</vertexbuffer>\n";
    
    ###
    ### VERTEX BUFFER NORMALS
    ###
    print FILEHANDLE "\t\t\t\t<vertexbuffer normals=\"true\">\n";
    foreach $vert(@$layervertnormals)
    {
      $work = $$layervertnormals{"$vert"};
      print FILEHANDLE "\t\t\t\t\t<vertex>\n";
      print FILEHANDLE "\t\t\t\t\t\t<normal $work />\n";
      print FILEHANDLE "\t\t\t\t\t</vertex>\n";
    }
    print FILEHANDLE "\t\t\t\t</vertexbuffer>\n";
    
    ###
    ### VERTEX BUFFER DIFFUSE COLORS
    ###
    print FILEHANDLE "\t\t\t\t<vertexbuffer colours_diffuse=\"false\">\n";
    print FILEHANDLE "\t\t\t\t</vertexbuffer>\n";
    
    ###
    ### VERTEX BUFFER TEXTURE COORDS
    ###
    print FILEHANDLE "\t\t\t\t<vertexbuffer texture_coord_dimensions_0=\"2\" texture_coord_dimensions_1=\"3\" texture_coords=\"2\">\n";
    foreach $vert(@$layerverts)
    {
      print FILEHANDLE "\t\t\t\t\t<vertex>\n";
      print FILEHANDLE "\t\t\t\t\t\t<texcoord u=\"0\" v=\"0\" />\n";
      print FILEHANDLE "\t\t\t\t\t\t<texcoord u=\"0\" v=\"0\" w=\"0\"/>\n";
      print FILEHANDLE "\t\t\t\t\t</vertex>\n";
    }
    print FILEHANDLE "\t\t\t\t</vertexbuffer>\n";
    
    print FILEHANDLE "\t\t\t</geometry>\n";
    print FILEHANDLE "\t\t</submesh>\n";
  }
  print FILEHANDLE "\t</submeshes>\n";
  print FILEHANDLE "</mesh>\n";

  # CLOSE THE FILE HANDLE
  close (FILEHANDLE);
}

######
###### SIMPLE XML PARSING ROUTINES, HAHA HACK HACK HACK
######

sub GetNextElement
{
  my $getNextElement = shift @_;  
  if ($getNextElement =~ m/[<]{1}([^>]+)[>]{1}/g)
  {
    return "<$1>";
  }
  # LuxPrint("No more elements.");
  return "";
}

#getAttribute, #element
sub GetAttribute
{
  my $getAttribute = shift @_;
  my $element = shift @_;
  
  #print "GetAttribute: [$getAttribute]\n";

  my @blob = split(/ /, $element);
  
  foreach my $innerBlob (@blob)
  {
    # LuxPrint("$innerBlob");
    my @nvp = split(/="/, $innerBlob);
    if (scalar(@nvp) > 1)
    {
      if ($nvp[0] eq $getAttribute)
      {
        my @valueEnd = split(/"/, $nvp[1]);
        if (scalar(@valueEnd) > 0)
        {
          return $valueEnd[0];
        }
      }
    }
  }
  
}

# $nestCount, $element
sub CalculateNestCount
{
  my $nestCount = shift @_;
  #LuxPrint("nestCount: $nestCount");
  
  my $element = shift @_;
  #LuxPrint("element: $element");
  
  if (index($element, "/>") != -1)
  {
  }
  elsif (index($element, "</") != -1)
  {
    $nestCount--;
  }
  else
  {
    $nestCount++;
  }
  return $nestCount;
}

# $nestCount
sub GetNestedWhitespace
{
  my $nestCount = shift @_;  
  #LuxPrint("nestCount: $nestCount");
  
  my $nestedWhitespace = "";
  for(my $index = 0; $index < $nestCount; ++$index)
  {
    $nestedWhitespace .= " ";
  }
  return $nestedWhitespace;
}

#element
sub GetElementName
{
  my $element = shift @_;

  my @splitElement = split("", $element);
  my $offset = 0;
  for(my $index = 0; $index < length($element); ++$index)
  {
    if ($splitElement[$index] eq "<")
    {
      ++$offset;
    }
    elsif ($splitElement[$index] eq "/")
    {
      if ($offset == 1)
      {
       ++$offset;
      }
      else
      {
        return substr($element, $offset, $index-$offset);
      }
    }
    elsif ($splitElement[$index] eq " ")
    {
      return substr($element, $offset, $index-$offset);
    }
    elsif ($splitElement[$index] eq ">")
    {
      return substr($element, $offset, $index-$offset);
    }
  }
  return $element;
}

# %simpleXPath, $nestCount, $element
sub CalculateXPath
{
  my $simpleXPath_ref = shift @_;
  #LuxPrint("simpleXPath: $simpleXPath_ref");
  #PrintHash($simpleXPath_ref);
  
  my $nestCount = shift @_;
  #LuxPrint("nestCount: $nestCount");
  
  my $element = shift @_;
  #LuxPrint("element: $element");
  
  my $elementName = GetElementName($element);
  $simpleXPath_ref->{$nestCount} = $elementName;
  #LuxPrint("Before Ret: " . $simpleXPath_ref->{$nestCount});
}

# %simpleXPath, $nestCount
sub GetXPath
{
  my $simpleXPath_ref = shift @_;
  #LuxPrint("simpleXPath_ref: $simpleXPath_ref");
  
  my $nestCount = shift @_;
  #LuxPrint("nestCount: $nestCount");
  
  my $ret = "";
  for(my $index = 0; $index <= $nestCount; ++$index)
  {
    if ($index > 0)
    {
      $ret .= "\\";
    }
    $ret .= $simpleXPath_ref->{$index};
  }
  return $ret;
}

# %uniqueXPaths, %simpleXPath, $nestCount
sub BuildUniqueXPaths
{
  my $uniqueXPaths_ref = shift @_;
  #LuxPrint("uniqueXPaths_ref: $uniqueXPaths_ref");
  
  my $simpleXPath_ref = shift @_;
  #LuxPrint("simpleXPath_ref: $simpleXPath_ref");
  
  my $nestCount = shift @_;
  #LuxPrint("nestCount: $nestCount");
  
  my $xpath = GetXPath($simpleXPath_ref, $nestCount);
  if (!defined($uniqueXPaths_ref->{$xpath}))
  {
    $uniqueXPaths_ref->{$xpath} = "";
    #LuxPrint("Added [$xpath]\n");
  }
}